﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Hexagonal_Minesweeper_2019
{
    public class GameData
    {
        public const int HEX_ROW_HEIGHT = 16;
        public const int HEX_COL_WIDTH = 12;

        private OptionsData m_optionsData;
        private StatisticsData m_statisticsData;
        private int m_numMinesCached = 0;

        public enum GamePlayStatus
        {
            NOT_STARTED,
            RUNNING,
            EXPLODED,
            WON
        }

        public DateTime startTime = DateTime.MinValue;
        public TimeSpan totalTimeNotCounted = TimeSpan.Zero;
        public int endTime = 0;
        public bool isNewBest = false;
        public GamePlayStatus gamePlayStatus;
        public Hex deathHex = null;
        public int minesLeft, unMinedHexesLeft;

        public enum HexStatus
        {
            UNCOVERED,
            COVERED,
            MARKED,
            QUESTION_MARKED
        };

        public class Hex
        {
            private Hex[,] m_arrayOwner;
            private int m_col, m_row;
            public int Col { get { return m_col; } }
            public int Row { get { return m_row; } }
            public HexStatus covered;
            public bool hasMine;

            public Hex (int c, int r, Hex[,] arrayOwner)
            {
                m_col = c;
                m_row = r;
                m_arrayOwner = arrayOwner;
                covered = HexStatus.COVERED;
                hasMine = false;
            }

            public List<Hex> GetNeighbors()
            {
                Array arr = m_arrayOwner;
                int hexesWide = arr.GetLength(0);
                int hexesHigh = arr.GetLength(1);

                List<Hex> retval = new List<Hex>();

                if (Row > 0)    // above
                    retval.Add(m_arrayOwner[Col, Row - 1]);
                if (Col > 0)
                    retval.Add(m_arrayOwner[Col - 1, Row]);
                if (Col < hexesWide - 1)
                    retval.Add(m_arrayOwner[Col + 1, Row]);
                if (Row < hexesHigh - 1)    // below
                    retval.Add(m_arrayOwner[Col, Row + 1]);
                if ((Col % 2) != 0 && Row < hexesHigh - 1)
                {
                    if (Col > 0)
                        retval.Add(m_arrayOwner[Col - 1, Row + 1]);
                    if (Col < hexesWide - 1)
                        retval.Add(m_arrayOwner[Col + 1, Row + 1]);
                }
                else if ((Col % 2) == 0 && Row > 0)
                {
                    if (Col > 0)
                        retval.Add(m_arrayOwner[Col - 1, Row - 1]);
                    if (Col < hexesWide - 1)
                        retval.Add(m_arrayOwner[Col + 1, Row - 1]);
                }
                return retval;
            }

            public void GetGridCoords (out int x, out int y)
            {
                y = Row * HEX_ROW_HEIGHT;
                x = Col * HEX_COL_WIDTH;
                if ((Col % 2) != 0)
                    y += HEX_ROW_HEIGHT / 2;
            }
        }
        private Hex[,] m_hexArray = null;

        public GameData (OptionsData options, StatisticsData statistics)
        {
            m_optionsData = options;
            m_statisticsData = statistics;
        }

        public OptionsData Options
        {
            get { return m_optionsData; }
        }

        public StatisticsData Statistics
        {
            get { return m_statisticsData; }
        }

        public int HexesWide
        {
            get
            {
                if (m_hexArray == null)
                    return 0;
                return m_hexArray.GetLength(0);
            }
        }
        public int HexesHigh
        {
            get
            {
                if (m_hexArray == null)
                    return 0;
                return m_hexArray.GetLength(1);
            }
        }
        public int NumMines
        {
            get { return m_numMinesCached; }
        }

        public void CreateHexArrayFromOptions()
        {
            m_hexArray = new Hex[Options.hexesWide, Options.hexesHigh];
            for (int i = 0; i < Options.hexesWide; ++i)
                for (int j = 0; j < Options.hexesHigh; ++j)
                    m_hexArray[i, j] = new Hex(i, j, m_hexArray);
            Random r = new Random();
            int col, row;
            for (int i = 0; i < Options.numMines; ++i)
            {
                do
                {
                    col = r.Next(0, Options.hexesWide);
                    row = r.Next(0, Options.hexesHigh);
                } while (m_hexArray[col, row].hasMine);
                m_hexArray[col, row].hasMine = true;
            }
            m_numMinesCached = Options.numMines;

            startTime = DateTime.MinValue;
            endTime = 0;
            totalTimeNotCounted = TimeSpan.Zero;
            isNewBest = false;
            gamePlayStatus = GamePlayStatus.NOT_STARTED;
            deathHex = null;
            minesLeft = m_numMinesCached;
            unMinedHexesLeft = Options.hexesWide * Options.hexesHigh - m_numMinesCached;
        }

        // We don't want to kill the player on the opening move.  In fact, we want
        // to ensure that the first hex the player clicks on will not only be
        // mine-free, but have no mines in any of the adjacent hexes either.
        public void RearrangeMinesAwayFromHex (Hex hex)
        {
            int numMinesToReDeploy = 0;
            List<Hex> hexAndNeighbors = hex.GetNeighbors();
            hexAndNeighbors.Add(hex);
            foreach (Hex h in hexAndNeighbors)
                if (h.hasMine)
                {
                    h.hasMine = false;
                    ++numMinesToReDeploy;
                }

            Random r = new Random();
            for (int i = 0; i < numMinesToReDeploy; ++i)
            {
                Hex h;
                do
                {
                    int col = r.Next(0, Options.hexesWide);
                    int row = r.Next(0, Options.hexesHigh);
                    h = m_hexArray[col, row];
                } while (h.hasMine  ||  hexAndNeighbors.Contains (h));
                h.hasMine = true;
            }
        }

        public Hex GetHex (int col, int row)
        {
            return m_hexArray[col, row];
        }

        public byte CountNeighborsWithMines (Hex hex)
        {
            byte n = 0;
            List<Hex> neighbors = hex.GetNeighbors();
            foreach (Hex h in neighbors)
                if (h.hasMine)
                    ++n;
            return n;
        }

        public byte CountUncoveredNeighbors (Hex hex)
        {
            byte n = 0;
            List<Hex> neighbors = hex.GetNeighbors();
            foreach (Hex h in neighbors)
                if (h.covered == HexStatus.UNCOVERED)
                    ++n;
            return n;
        }

        public int ElapsedTime
        {
            get
            {
                if (endTime > 0)
                    return endTime;
                int retval = (int)(DateTime.Now - startTime - totalTimeNotCounted).TotalSeconds;
                if (retval > 999)
                    retval = 999;
                return retval;
            }
        }

        public bool UncoverHex (Hex hex)
        {
            if (hex.covered == HexStatus.COVERED)
            {
                if (gamePlayStatus == GamePlayStatus.NOT_STARTED)
                {
                    RearrangeMinesAwayFromHex(hex);
                    gamePlayStatus = GamePlayStatus.RUNNING;
                    startTime = DateTime.Now;
                }

                hex.covered = HexStatus.UNCOVERED;
                if (hex.hasMine)
                    BlowUp(hex);
                if (gamePlayStatus == GamePlayStatus.EXPLODED)
                    return false;
                if (--unMinedHexesLeft <= 0  ||  Options.rightToLeft  &&  hex.Col == 0)
                {
                    // Hooray!
                    endTime = ElapsedTime;
                    gamePlayStatus = GamePlayStatus.WON;
                    foreach (Hex h in m_hexArray)
                        if (h.covered == HexStatus.COVERED && h.hasMine)
                        {
                            h.covered = HexStatus.MARKED;
                            --minesLeft;
                        }
                        else if (h.covered == HexStatus.MARKED && !h.hasMine)
                        {
                            h.covered = HexStatus.COVERED;
                            ++minesLeft;
                        }
                    StatisticsData.StatisicsForOneDifficulty sfod = Statistics.FindByDifficulty
                        (NumMines, HexesWide, HexesHigh, Options.rightToLeft);
                    if (sfod == null)
                    {
                        sfod = new StatisticsData.StatisicsForOneDifficulty(NumMines, HexesWide, HexesHigh, Options.rightToLeft);
                        Statistics.statCollection.Add(sfod);
                    }
                    if (sfod.DoGameWonInTime(ElapsedTime))
                        isNewBest = true;
                    Statistics.Serialize();
                }
                else if (CountNeighborsWithMines(hex) == 0)
                    return ClearAroundHex(hex);
            }
            return true;
        }

        private void BlowUp(Hex hex)
        {
            endTime = ElapsedTime;
            gamePlayStatus = GamePlayStatus.EXPLODED;
            deathHex = hex;
            foreach (Hex h in m_hexArray)
                if (h.hasMine &&  h.covered != HexStatus.MARKED)
                    h.covered = HexStatus.UNCOVERED;
            StatisticsData.StatisicsForOneDifficulty sfod = Statistics.FindByDifficulty
                (NumMines, HexesWide, HexesHigh, Options.rightToLeft);
            if (sfod == null)
            {
                sfod = new StatisticsData.StatisicsForOneDifficulty(NumMines, HexesWide, HexesHigh, Options.rightToLeft);
                Statistics.statCollection.Add(sfod);
            }
            ++ sfod.gamesPlayed;
            if (sfod.currentStreak < 0)
                --sfod.currentStreak;
            else
                sfod.currentStreak = -1;
            if (-sfod.currentStreak > sfod.longestLoseStreak)
                sfod.longestLoseStreak = -sfod.currentStreak;
            Statistics.Serialize();
        }

        private byte CountMarkedNeighbors (Hex hex)
        {
            byte n = 0;
            List<Hex> neighbors = hex.GetNeighbors();
            foreach (Hex h in neighbors)
                if (h.covered == HexStatus.MARKED)
                    ++n;
            return n;
        }

        public List<Hex> NeighboringCoveredUnmarkedHexes (Hex hex)
        {
            List<Hex> retval = new List<Hex>();
            List<Hex> neighbors = hex.GetNeighbors();
            foreach (Hex h in neighbors)
                if (h.covered == HexStatus.COVERED)
                    retval.Add(h);
            return retval;
        }

        public bool ClearAroundHex (Hex hex)
        {
            byte adjacentCount = CountNeighborsWithMines (hex);
            if (adjacentCount > 0  &&  CountMarkedNeighbors(hex) != adjacentCount)
                return true;
            List<Hex> neighbors = hex.GetNeighbors();
            foreach (Hex h in neighbors)
            {
                if (!UncoverHex(h))
                    return false;
                else if (gamePlayStatus == GamePlayStatus.WON)
                    return true;
            }
            return true;
        }

        public Hex FindHex (int x, int y)
        {
            int col = x / HEX_COL_WIDTH;
            // Each odd-numbered column is shifted half a row height downward
            if (col % 2 != 0)
                y -= HEX_ROW_HEIGHT / 2;
            int row = y / HEX_ROW_HEIGHT;
            if (col < 0 || col >= HexesWide || row < 0 || row >= HexesHigh)
                return null;
            return m_hexArray[col, row];
        }
    }
}
