﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Serialization;

namespace Hexagonal_Minesweeper_2019
{
    public class StatisticsData
    {
        public class StatisicsForOneDifficulty
        {
            // These four values uniquely identify each difficulty:
            public int Mines { get; set; }
            public int Columns { get; set; }
            public int Rows { get; set; }
            public bool RightToLeft { get; set; }

            public const int NUM_BEST_TIMES = 5;

            public int[] bestTimes = new int[NUM_BEST_TIMES];
            public DateTime[] dateOfBestTimes = new DateTime[NUM_BEST_TIMES];

            public int gamesPlayed = 0;
            public int gamesWon = 0;
            public int longestWinStreak = 0;
            public int longestLoseStreak = 0;
            public int currentStreak = 0;   // a negative value indicates a losing streak

            public int WinPercentage
            {
                get
                {
                    if (gamesPlayed <= 0)
                        return 0;
                    return (gamesWon * 100) / gamesPlayed;
                }
            }

            public StatisicsForOneDifficulty()
            {
                // Pre-populate with invalid values
                Mines = 0;
                Columns = 0;
                Rows = 0;
                RightToLeft = false;
            }

            public StatisicsForOneDifficulty (int mines, int columns, int rows, bool isrightToLeft)
            {
                Mines = mines;
                Columns = columns;
                Rows = rows;
                RightToLeft = isrightToLeft;

                for (int i = 0; i < NUM_BEST_TIMES; ++i)
                {
                    bestTimes[i] = 999;
                    dateOfBestTimes[i] = DateTime.MinValue;
                }
            }

            // Returns true if this game is the new best.
            public bool DoGameWonInTime (int elapsedTime)
            {
                ++gamesPlayed;
                ++gamesWon;
                if (currentStreak > 0)
                    ++currentStreak;
                else
                    currentStreak = 1;
                if (currentStreak > longestWinStreak)
                    longestWinStreak = currentStreak;
                
                for (int i = 0; i < NUM_BEST_TIMES; ++i)
                {
                    if (elapsedTime < bestTimes[i])
                    {
                        // Move this and all all succeeding best times down by one.
                        for (int j = NUM_BEST_TIMES - 1; j > i; --j)
                        {
                            bestTimes[j] = bestTimes[j-1];
                            dateOfBestTimes[j] = dateOfBestTimes[j-1];
                        }
                        bestTimes[i] = elapsedTime;
                        dateOfBestTimes[i] = DateTime.Now;
                        return i == 0;
                    }
                }
                return false;
            }

            public string DifficultyName
            {
                get
                {
                    string retval;
                    OptionsData.Difficulty diff = OptionsData.GetDifficultyFromValues(Mines, Columns, Rows);
                    switch (diff)
                    {
                        case OptionsData.Difficulty.BEGINNER:
                            retval = "Beginner";
                            break;
                        case OptionsData.Difficulty.INTERMEDIATE:
                            retval = "Intermediate";
                            break;
                        case OptionsData.Difficulty.EXPERT:
                            retval = "Expert";
                            break;
                        default:
                            retval = Mines + " mines, " + Columns + "x" + Rows;
                            break;
                    }
                    if (RightToLeft)
                        retval += " (right-to-left)";
                    return retval;
                }
            }
        }

        public List<StatisicsForOneDifficulty> statCollection = new List<StatisicsForOneDifficulty>();

        public StatisticsData()
        {
        }

        public StatisticsData (bool prepopulate)
        {
            if (prepopulate)
            {
                statCollection.Add(new StatisicsForOneDifficulty(15, OptionsData.MIN_HEXES_WIDE, OptionsData.MIN_HEXES_HIGH, false));
                statCollection.Add(new StatisicsForOneDifficulty(60, 24, 16, false));
                statCollection.Add(new StatisicsForOneDifficulty(135, 36, 16, false));
            }
        }

        public StatisicsForOneDifficulty FindByDifficulty (int mines, int columns, int rows, bool rightToLeft)
        {
            foreach (StatisicsForOneDifficulty sfod in statCollection)
            {
                if (sfod.Mines == mines  &&  sfod.Columns == columns  &&
                    sfod.Rows == rows && sfod.RightToLeft == rightToLeft)
                        return sfod;
            }
            return null;
        }

        public void ResetAll()
        {
            foreach (StatisicsForOneDifficulty sfod in statCollection)
            {
                sfod.gamesPlayed = sfod.gamesWon = 0;
                sfod.longestWinStreak = sfod.longestLoseStreak = sfod.currentStreak = 0;
                for (int i = 0; i < StatisicsForOneDifficulty.NUM_BEST_TIMES; ++i)
                {
                    sfod.bestTimes[i] = 999;
                    sfod.dateOfBestTimes[i] = DateTime.MinValue;
                }
            }
            Serialize();
        }

        public bool Serialize()
        {
            FileStream stream = null;
            try
            {
                string dir = MainWindow.AppDataDir;
                if (!Directory.Exists(dir))
                    Directory.CreateDirectory(dir);
                string filename = Path.Combine(dir, "statistics.xml");
                stream = new FileStream(filename, FileMode.Create, FileAccess.Write, FileShare.None);
                XmlSerializer writer = new XmlSerializer(typeof(StatisticsData));
                writer.Serialize(stream, this);
                stream.Close();
                return true;
            }
            catch
            {
                if (stream != null)
                    stream.Close();
                return false;
            }
        }

        public static StatisticsData Deserialize()
        {
            try
            {
                string dir = MainWindow.AppDataDir;
                if (!Directory.Exists(dir))
                    return null;

                string filename = Path.Combine(dir, "statistics.xml");
                FileStream stream = new FileStream(filename, FileMode.Open, FileAccess.Read, FileShare.None);
                XmlSerializer reader = new XmlSerializer(typeof(StatisticsData));
                StatisticsData data = reader.Deserialize(stream) as StatisticsData;
                stream.Close();
                return data;
            }
            catch
            {
                return null;
            }
        }
    }
}
